import tqdm
import stempeg
import argparse
from pathlib import Path
from musdb import DB
import sys


def musdb_convert(inargs=None):
    """
    cli application to conver audio files to spectrogram images
    """
    parser = argparse.ArgumentParser()

    parser.add_argument(
        'musdb_root',
        type=str,
    )

    parser.add_argument(
        'output_root',
        type=str
    )

    parser.add_argument(
        '--download', action='store_true', default=False,
        help='use the downloaded 7s musdb stems',
    )

    parser.add_argument(
        '--extension', type=str, default='.wav'
    )

    args = parser.parse_args(inargs)

    mus = DB(root=args.musdb_root, download=args.download)

    for track in tqdm.tqdm(mus):

        track_estimate_dir = Path(
            args.output_root, track.subset, track.name
        )
        track_estimate_dir.mkdir(exist_ok=True, parents=True)
        # write out tracks to disk

        stempeg.write_audio(
            path=str(track_estimate_dir / Path('mixture').with_suffix(args.extension)),
            data=track.audio,
            sample_rate=track.rate
        )
        for name, track in track.targets.items():
            stempeg.write_audio(
                path=str(track_estimate_dir / Path(name).with_suffix(args.extension)),
                data=track.audio,
                sample_rate=track.rate
            )


if __name__ == '__main__':
    musdb_convert(sys.argv[1:])



import pytest
import musdb.audio_classes as ac
import musdb
import numpy as np


@pytest.fixture(params=[True, False])
def mus(request):
    return musdb.DB(root='data/MUS-STEMS-SAMPLE', is_wav=request.param)


@pytest.fixture(params=[0, 1, 2, 11.1, None])
def chunk_duration(request):
    return request.param


@pytest.fixture(params=[44100.0, 22050.0, 16000.0])
def sample_rate(request):
    return request.param


def test_targets(mus):
    for track in mus:
        for key, target in list(track.targets.items()):
            assert target.audio.shape


def test_rates(mus):
    for track in mus:
        assert track.rate == 44100
        assert track.audio.shape
        for key, source in list(track.sources.items()):
            assert source.rate == 44100
            assert source.audio.shape


def test_durations(mus):
    for track in mus:
        assert track.duration > 0
        # check if duration is with zero pad range of mp4 files
        # (one frame is typically 1024 samples)
        assert track.audio.shape[0] > (track.duration * track.rate - 1024) \
            and track.audio.shape[0] < (track.duration * track.rate + 1024)


def test_sample_rate(mus, sample_rate):
    track = mus[0]
    track.sample_rate = sample_rate
    track.chunk_duration = 1.0
    assert track.audio.shape[0] == sample_rate


def test_chunking(mus, chunk_duration):
    for track in mus:
        track.chunk_duration = chunk_duration
        if chunk_duration:
            assert np.allclose(
                track.audio.shape[0], chunk_duration * track.rate)
            for _, target in track.sources.items():
                assert np.allclose(
                    target.audio.shape[0], chunk_duration * track.rate)
            for _, target in track.targets.items():
                assert np.allclose(
                    target.audio.shape[0], chunk_duration * track.rate)


def test_track_artisttitle(mus):
    track = ac.MultiTrack(name="None", path="None")

    assert track.artist is None
    assert track.title is None


def test_source(mus):
    mtrack = ac.MultiTrack(name="abc123", path="None")
    with pytest.raises(ValueError):
        source = ac.Source(mtrack, name="test", path="None")
        source.audio

    source.audio = np.zeros((2, 44100))
    assert source.audio.shape == (2, 44100)


def test_track(mus):
    with pytest.raises(ValueError):
        track = ac.Track(path="None")
        track.audio

    track.audio = np.zeros((2, 44100))
    assert track.audio.shape == (2, 44100)



import os
import numpy as np
import stempeg


class Track(object):
    """
    Generic audio Track that can be wav or stem file

    Attributes
    ----------
    name : str
        Track name
    path : str
        Absolute path of mixture audio file
    stem_id : int
        stem/substream ID
    is_wav : boolean
        If stem is read from wav or mp4 stem
    subset : {'train', 'test'}
        Track belongs to which subset.
    targets : OrderedDict
        OrderedDict of mixted Targets for this ``track``.
    sources : Dict
        Dict of ``Source`` objects for this ``track``.
    chunk_start : float
        sets offset when loading the audio, defaults to 0 (beginning).
    chunk_duration : float
        sets duration for the audio, defaults to ``None`` (end).
    """

    def __init__(
        self,
        path="None",
        is_wav=False,
        stem_id=None,
        subset=None,
        chunk_start=0,
        chunk_duration=None,
        sample_rate=None
    ):
        self.path = path
        self.subset = subset
        self.stem_id = stem_id
        self.is_wav = is_wav
        self.chunk_start = chunk_start
        self.chunk_duration = chunk_duration
        self.sample_rate = sample_rate

        # load and store metadata
        if os.path.exists(self.path):
            self.info = stempeg.Info(self.path)
            self.samples = int(self.info.samples(self.stem_id))
            self.duration = self.info.duration(self.stem_id)
            self.rate = self.info.rate(self.stem_id)
        else:
            # set to `None` if no path was set (fake file)
            self.info = None
            self.samples = None
            self.duration = None
            self.rate = None

        self._audio = None

    def __len__(self):
        return self.samples

    @property
    def audio(self):
        # return cached audio if explicitly set by setter
        if self._audio is not None:
            return self._audio
        # read from disk to save RAM otherwise
        else:
            return self.load_audio(
                self.path,
                self.stem_id,
                self.chunk_start,
                self.chunk_duration,
                self.sample_rate
            )

    @audio.setter
    def audio(self, array):
        self._audio = array

    def load_audio(
        self,
        path,
        stem_id,
        chunk_start=0,
        chunk_duration=None,
        sample_rate=None
    ):
        """array_like: [shape=(num_samples, num_channels)]
        """
        if os.path.exists(self.path):
            if self.is_wav:
                stem_id = 0
            audio, rate = stempeg.read_stems(
                filename=path,
                stem_id=stem_id,
                start=chunk_start,
                duration=chunk_duration,
                info=self.info,
                sample_rate=sample_rate,
                ffmpeg_format="s16le"
            )
            self._rate = rate
            return audio
        else:
            self._rate = None
            self._audio = None
            raise ValueError("Oops! %s cannot be loaded" % path)

    def __repr__(self):
        return "%s" % (self.path)


class MultiTrack(Track):
    def __init__(
        self,
        path=None,
        name=None,
        artist=None,
        title=None,
        sources=None,
        targets=None,
        sample_rate=None,
        *args,
        **kwargs
    ):
        super(MultiTrack, self).__init__(path=path, *args, **kwargs)

        self.name = name
        self.path = path
        try:
            split_name = name.split(' - ')
            self.artist = split_name[0]
            self.title = split_name[1]
        except IndexError:
            self.artist = artist
            self.title = title

        self.sources = sources
        self.targets = targets
        self.sample_rate = sample_rate
        self._stems = None

    @property
    def stems(self):
        """array_like: [shape=(stems, num_samples, num_channels)]
        """

        # return cached audio it explicitly set bet setter
        if self._stems is not None:
            return self._stems
        # read from disk to save RAM otherwise
        else:
            if not self.is_wav and os.path.exists(self.path):
                S, rate = stempeg.read_stems(
                    filename=self.path,
                    start=self.chunk_start,
                    duration=self.chunk_duration,
                    info=self.info,
                    sample_rate=self.sample_rate,
                    ffmpeg_format="s16le"
                )
            else:
                rate = self.rate
                S = []
                S.append(self.audio)
                # append sources in order of stem_ids
                for k, v in sorted(self.sources.items(), key=lambda x: x[1].stem_id):
                    S.append(v.audio)
                S = np.array(S)
            self._rate = rate
            return S

    def __repr__(self):
        return "%s" % (self.name)

class Source(Track):
    """
    An audio Target which is a linear mixture of several sources

    Attributes
    ----------
    name : str
        Name of this source
    stem_id : int
        stem/substream ID is set here.
    is_wav : boolean
        If stem is read from wav or mp4 stem
    path : str
        Absolute path to audio file
    gain : float
        Mixing weight for this source
    """
    def __init__(
        self,
        multitrack,    # belongs to a multitrack
        name=None,     # has its own name
        path=None,     # might have its own path
        stem_id=None,  # might have its own stem_id
        gain=1.0,
        *args,
        **kwargs
    ):
        self.multitrack = multitrack
        self.name = name
        self.path = path
        self.stem_id = stem_id
        self.gain = gain
        self._audio = None

    def __repr__(self):
        return self.path

    @property
    def audio(self):
        # return cached audio if explicitly set by setter
        if self._audio is not None:
            return self._audio
        # read from disk to save RAM otherwise
        else:
            return self.multitrack.load_audio(
                self.path,
                self.stem_id,
                self.multitrack.chunk_start,
                self.multitrack.chunk_duration,
                self.multitrack.sample_rate
            )

    @audio.setter
    def audio(self, array):
        self._audio = array

    @property
    def rate(self):
        return self.multitrack.rate

# Target Track from musdb DB mixed from several musdb Tracks
class Target(Track):
    """
    An audio Target which is a linear mixture of several sources

    Attributes
    ----------
    multitrack : Track
        Track object
    sources : list[Source]
        list of ``Source`` objects for this ``Target``
    """
    def __init__(
        self, 
        multitrack,
        sources,
        name=None,  # has its own name
    ):
        self.multitrack = multitrack
        self.sources = sources
        self.name = name

    @property
    def audio(self):
        """array_like: [shape=(num_samples, num_channels)]

        mixes audio for targets on the fly
        """
        mix_list = []
        for source in self.sources:
            audio = source.audio
            if audio is not None:
                mix_list.append(
                    source.gain * audio
                )
        return np.sum(np.array(mix_list), axis=0)

    @property
    def rate(self):
        return self.multitrack.rate

    def __repr__(self):
        parts = []
        for source in self.sources:
            parts.append(source.name)
        return '+'.join(parts)



from .audio_classes import MultiTrack, Source, Target
from os import path as op
from pathlib import Path
import stempeg
from urllib.request import urlopen, Request
import collections
import numpy as np
from tqdm import tqdm
import zipfile
import yaml
import errno
import musdb
import shutil
import os
import tempfile


class DB(object):
    """
    The musdb DB Object

    Parameters
    ----------
    root : str, optional
        musdb Root path. If set to `None` it will be read
        from the `MUSDB_PATH` environment variable

    subsets : str or list, optional
        select a _musdb_ subset `train` or `test` (defaults to both)

    is_wav : boolean, optional
        expect subfolder with wav files for each source instead stems,
        defaults to `False`

    download : boolean, optional
        download sample version of MUSDB18 which includes 7s excerpts,
        defaults to `False`

    subsets : list[str], optional
        select a _musdb_ subset `train` or `test`.
        Default `None` loads `['train', 'test']`

    split : str, optional
        when `subsets=train`, `split` selects the train/validation split.
        `split='train' loads the training split, `split='valid'` loads the validation
        split. `split=None` applies no splitting.

    Attributes
    ----------
    setup_file : str
        path to yaml file. default: `setup.yaml`
    root : str
        musdb Root path. Default is `MUSDB_PATH`. In combination with
        `download`, this path will set the download destination and set to
        '~/musdb/' by default.
    sources_dir : str
        path to Sources directory
    sources_names : list[str]
        list of names of available sources
    targets_names : list[str]
        list of names of available targets
    setup : Dict
        loaded yaml configuration
    sample_rate : Optional(Float)
        sets sample rate for optional resampling. Defaults to none
        which results in `44100.0`

    Methods
    -------
    load_mus_tracks()
        Iterates through the musdb folder structure and
        returns ``Track`` objects

    """

    def __init__(
        self,
        root=None,
        setup_file=None,
        is_wav=False,
        download=False,
        subsets=["train", "test"],
        split=None,
        sample_rate=None,
    ):
        if root is None:
            if download:
                self.root = os.path.expanduser("~/MUSDB18/MUSDB18-7")
            else:
                if "MUSDB_PATH" in os.environ:
                    self.root = os.environ["MUSDB_PATH"]
                else:
                    raise RuntimeError("Variable `MUSDB_PATH` has not been set.")
        else:
            self.root = os.path.expanduser(root)

        if setup_file is not None:
            setup_path = op.join(self.root, setup_file)
        else:
            setup_path = os.path.join(musdb.__path__[0], "configs", "mus.yaml")

        with open(setup_path, "r") as f:
            self.setup = yaml.safe_load(f)

        if download:
            self.url = self.setup["sample-url"]
            self.download()
            if not self._check_exists():
                raise RuntimeError(
                    "Dataset not found."
                    + "You can use download=True to download a sample version of the dataset"
                )

        self.sample_rate = sample_rate
        self.sources_names = list(self.setup["sources"].keys())
        self.targets_names = list(self.setup["targets"].keys())
        self.is_wav = is_wav
        self.tracks = self.load_mus_tracks(subsets=subsets, split=split)

    def __getitem__(self, index):
        return self.tracks[index]

    def __len__(self):
        return len(self.tracks)

    def get_validation_track_indices(self, validation_track_names=None):
        """Returns validation track indices by a given list of track names

        Defaults to the builtin selection 8 validation tracks, defined in
        `mus.yaml`.

        Parameters
        == == == == ==
        validation_track_names : list[str], optional
            validation track names by a given `str` or list of tracknames

        Returns
        -------
        list[int]
            return a list of validation track indices
        """
        if validation_track_names is None:
            validation_track_names = self.setup["validation_tracks"]

        return self.get_track_indices_by_names(validation_track_names)

    def get_track_indices_by_names(self, names):
        """Returns musdb track indices by track name

        Can be used to filter the musdb tracks for
        a validation subset by trackname

        Parameters
        == == == == ==
        names : list[str], optional
            select tracks by a given `str` or list of tracknames

        Returns
        -------
        list[int]
            return a list of ``Track`` Objects
        """
        if isinstance(names, str):
            names = [names]

        return [[t.name for t in self.tracks].index(name) for name in names]

    def load_mus_tracks(self, subsets=None, split=None):
        """Parses the musdb folder structure, returns list of `Track` objects

        Parameters
        ==========
        subsets : list[str], optional
            select a _musdb_ subset `train` or `test`.
            Default `None` loads [`train, test`].
        split : str
            for subsets='train', `split='train` applies a train/validation split.
            if `split='valid`' the validation split of the training subset will be used


        Returns
        -------
        list[Track]
            return a list of ``Track`` Objects
        """

        if subsets is not None:
            if isinstance(subsets, str):
                subsets = [subsets]
        else:
            subsets = ["train", "test"]

        if subsets != ["train"] and split is not None:
            raise RuntimeError("Subset has to set to `train` when split is used")

        tracks = []
        for subset in subsets:
            subset_folder = op.join(self.root, subset)

            for _, folders, files in os.walk(subset_folder):
                if self.is_wav:
                    # parse pcm tracks and sort by name
                    for track_name in sorted(folders):
                        if subset == "train":
                            if (
                                split == "train"
                                and track_name in self.setup["validation_tracks"]
                            ):
                                continue
                            elif (
                                split == "valid"
                                and track_name not in self.setup["validation_tracks"]
                            ):
                                continue

                        track_folder = op.join(subset_folder, track_name)
                        # create new mus track
                        track = MultiTrack(
                            name=track_name,
                            path=op.join(track_folder, self.setup["mixture"]),
                            subset=subset,
                            is_wav=self.is_wav,
                            stem_id=self.setup["stem_ids"]["mixture"],
                            sample_rate=self.sample_rate,
                        )

                        # add sources to track
                        sources = {}
                        for src, source_file in list(self.setup["sources"].items()):
                            # create source object
                            abs_path = op.join(track_folder, source_file)
                            if os.path.exists(abs_path):
                                sources[src] = Source(
                                    track,
                                    name=src,
                                    path=abs_path,
                                    stem_id=self.setup["stem_ids"][src],
                                    sample_rate=self.sample_rate,
                                )
                        track.sources = sources
                        track.targets = self.create_targets(track)

                        # add track to list of tracks
                        tracks.append(track)
                else:
                    # parse stem files
                    for track_name in sorted(files):
                        if not track_name.endswith(".stem.mp4"):
                            continue
                        if subset == "train":
                            if (
                                split == "train"
                                and track_name.split(".stem.mp4")[0]
                                in self.setup["validation_tracks"]
                            ):
                                continue
                            elif (
                                split == "valid"
                                and track_name.split(".stem.mp4")[0]
                                not in self.setup["validation_tracks"]
                            ):
                                continue

                        # create new mus track
                        track = MultiTrack(
                            name=track_name.split(".stem.mp4")[0],
                            path=op.join(subset_folder, track_name),
                            subset=subset,
                            stem_id=self.setup["stem_ids"]["mixture"],
                            is_wav=self.is_wav,
                            sample_rate=self.sample_rate,
                        )
                        # add sources to track
                        sources = {}
                        for src, source_file in list(self.setup["sources"].items()):
                            # create source object
                            abs_path = op.join(subset_folder, track_name)
                            if os.path.exists(abs_path):
                                sources[src] = Source(
                                    track,
                                    name=src,
                                    path=abs_path,
                                    stem_id=self.setup["stem_ids"][src],
                                    sample_rate=self.sample_rate,
                                )
                        track.sources = sources

                        # add targets to track
                        track.targets = self.create_targets(track)
                        tracks.append(track)

        return tracks

    def create_targets(self, track):
        # add targets to track
        targets = collections.OrderedDict()
        for name, target_srcs in list(self.setup["targets"].items()):
            # add a list of target sources
            target_sources = []
            for source, gain in list(target_srcs.items()):
                if source in list(track.sources.keys()):
                    # add gain to source tracks
                    track.sources[source].gain = float(gain)
                    # add tracks to components
                    target_sources.append(track.sources[source])
                    # add sources to target
            if target_sources:
                targets[name] = Target(track, sources=target_sources, name=name)

        return targets

    def save_estimates(self, user_estimates, track, estimates_dir, write_stems=False):
        """Writes `user_estimates` to disk while recreating the musdb file structure in that folder.

        Parameters
        ==========
        user_estimates : Dict[np.array]
            the target estimates.
        track : Track,
            musdb track object
        estimates_dir : str,
            output folder name where to save the estimates.
        """
        track_estimate_dir = op.join(estimates_dir, track.subset, track.name)
        if not os.path.exists(track_estimate_dir):
            os.makedirs(track_estimate_dir)

        # write out tracks to disk
        if write_stems:
            pass
            # to be implemented
        else:
            for target, estimate in list(user_estimates.items()):
                target_path = op.join(track_estimate_dir, target + ".wav")
                stempeg.write_audio(
                    path=target_path, data=estimate, sample_rate=track.rate
                )

    def _check_exists(self):
        return os.path.exists(os.path.join(self.root, "train"))

    def download(self, progress: bool = True, suffix: str = ".zip"):
        """Download the MUSDB Sample data"""
        if self._check_exists():
            return

        # download files
        try:
            os.makedirs(os.path.join(self.root))
        except OSError as e:
            if e.errno == errno.EEXIST:
                pass
            else:
                raise
        print("Downloading MUSDB 7s Sample Dataset to %s..." % self.root)

        file_size = None
        req = Request(self.url, headers={"User-Agent": "musdb_downloader"})
        u = urlopen(req)
        meta = u.info()
        if hasattr(meta, "getheaders"):
            content_length = meta.getheaders("Content-Length")
        else:
            content_length = meta.get_all("Content-Length")
        if content_length is not None and len(content_length) > 0:
            file_size = int(content_length[0])

        # We deliberately save it in a temp file and move it after
        # download is complete. This prevents a local working checkpoint
        # being overridden by a broken download.
        f = tempfile.NamedTemporaryFile(delete=False, suffix=suffix)

        try:
            with tqdm(
                total=file_size,
                disable=not progress,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            ) as pbar:
                while True:
                    buffer = u.read(8192)
                    if len(buffer) == 0:
                        break
                    f.write(buffer)
                    pbar.update(len(buffer))

            f.close()
            zip_ref = zipfile.ZipFile(f.name, "r")
            zip_ref.extractall(os.path.join(self.root))
            zip_ref.close()
        finally:
            f.close()
            if os.path.exists(f.name):
                os.remove(f.name)



import os
import pytest
import musdb
import numpy as np
import yaml


@pytest.fixture(params=['train', 'test', ['train', 'test'], None])
def subset(request):
    return request.param


@pytest.fixture(params=[True, False])
def mus(request, subset):
    return musdb.DB(
        root='data/MUS-STEMS-SAMPLE', is_wav=request.param, subsets=subset
    )


def test_stems(mus):
    setup_path = os.path.join(
        musdb.__path__[0], 'configs', 'mus.yaml'
    )

    with open(setup_path, 'r') as f:
        setup = yaml.safe_load(f)

    for track in mus:
        for k, v in setup['stem_ids'].items():
            if k == 'mixture':
                assert np.allclose(
                    track.audio,
                    track.stems[v]
                )
            else:
                assert np.allclose(
                    track.sources[k].audio,
                    track.stems[v]
                )


def test_file_loading(mus, subset):
    for track in mus:
        assert track.audio.shape[1] > 0
        assert track.audio.shape[-1] == 2
        assert track.stems.shape[0] == 5

    # loads only the train set
    if subset == 'train':
        assert len(mus) == 1

    # load train and test set
    if subset == ['train', 'test']:
        assert len(mus) == 2

    # load train and test set
    if subset is None:
        assert len(mus) == 2


def test_audio_regression():
    """test audio loading capabilities"""
    mus = musdb.DB(download=True)
    s = 0
    for track in mus:
        s += track.audio.sum()

    assert np.allclose(s, -24778.126983642578)


def test_download_and_validation():
    mus_all = musdb.DB(download=True)

    assert len(mus_all) == 144

    for track in mus_all:
        assert track.audio.shape[1] > 0
        assert track.audio.shape[-1] == 2
        assert track.stems.shape[0] == 5

    mus_test = musdb.DB(download=True, subsets='test', split=None)
    assert len(mus_test) == 50

    mus_train = musdb.DB(download=True, subsets='train', split='train')
    assert len(mus_train) == 80
    # test validation set
    mus_valid = musdb.DB(download=True, subsets='train', split='valid')
    assert len(mus_valid) == 14

    assert len(mus_train) == len(mus_all) - len(mus_test) - len(mus_valid)

    with pytest.raises(RuntimeError):
        mus_train = musdb.DB(download=True, split='train')



import setuptools
from importlib.machinery import SourceFileLoader


version = SourceFileLoader(
    'musdb.version', 'musdb/version.py'
).load_module()

with open('README.md', 'r') as fdesc:
    long_description = fdesc.read()


if __name__ == "__main__":
    setuptools.setup(
        # Name of the project
        name='musdb',

        # Version
        version=version.version,

        # Description
        description='Python parser for the SIGSEP MUSDB18 dataset',
        url='https://github.com/sigsep/sigsep-mus-db',

        # Your contact information
        author='Fabian-Robert Stoeter',
        author_email='mail@faroit.com',

        # License
        license='MIT',

        # Packages in this project
        # find_packages() finds all these automatically for you
        packages=setuptools.find_packages(),

        long_description=long_description,
        long_description_content_type='text/markdown',

        # Dependencies, this installs the entire Python scientific
        # computations stack
        install_requires=[
            'numpy>=1.7',
            'stempeg>=0.2.3',
            'pyaml',
            'tqdm'
        ],

        package_data={
            'musdb': ['configs/mus.yaml'],
        },

        extras_require={  # Optional
            'dev': ['check-manifest'],
            'tests': ['pytest'],
            'docs': [
                'sphinx',
                'sphinx_rtd_theme',
                'recommonmark'
            ],
        },

        tests_require=[
            'pytest'
        ],

        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Environment :: Plugins',
            'Intended Audience :: Telecommunications Industry',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Topic :: Multimedia :: Sound/Audio :: Analysis',
            'Topic :: Multimedia :: Sound/Audio :: Sound Synthesis'
        ],

        entry_points={
            'console_scripts': [
                'musdbconvert=musdb.tools:musdb_convert',
            ],
        },

        zip_safe=False,
    )
